# cython: language_level=3
# cython: binding=True

from cpython cimport pythread

include "_lock.pxi"


cdef class FastRLock:
    """Fast, re-entrant locking.

    Under non-congested conditions, the lock is never acquired but only
    counted.  Only when a second thread comes in and notices that the
    lock is needed, it acquires the lock and notifies the first thread
    to release it when it's done.  This is all made possible by the
    wonderful GIL.
    """
    cdef _LockStatus _real_lock

    def __cinit__(self):
        self._real_lock = _LockStatus(
            lock=pythread.PyThread_allocate_lock(),
            owner=0, is_locked=False, pending_requests=0, entry_count=0)
        if not self._real_lock.lock:
            raise MemoryError()

    def __dealloc__(self):
        if self._real_lock.lock:
            pythread.PyThread_free_lock(self._real_lock.lock)
            self._real_lock.lock = NULL

    # compatibility with RLock and expected Python level interface

    def acquire(self, bint blocking=True):
        return _lock_rlock(
            &self._real_lock, pythread.PyThread_get_thread_ident(), blocking)

    def release(self):
        if self._real_lock.entry_count == 0:
            raise RuntimeError("cannot release un-acquired lock")
        _unlock_lock(&self._real_lock)

    def __enter__(self):
        # self.acquire()
        if not _lock_rlock(
                &self._real_lock, pythread.PyThread_get_thread_ident(), blocking=True):
            raise LockNotAcquired()

    def __exit__(self, t, v, tb):
        # self.release()
        if self._real_lock.entry_count == 0 or self._real_lock.owner != pythread.PyThread_get_thread_ident():
            raise RuntimeError("cannot release un-acquired lock")
        _unlock_lock(&self._real_lock)

    def _is_owned(self):
        return self._real_lock.entry_count > 0 and self._real_lock.owner == pythread.PyThread_get_thread_ident()


cdef inline bint _lock_rlock(_LockStatus *lock, pythread_t current_thread,
                             bint blocking) nogil except -1:
    # Note that this function *must* hold the GIL when being called.
    # We just use 'nogil' in the signature to make sure that no Python
    # code execution slips in that might free the GIL

    if lock.entry_count:
        # locked! - by myself?
        if lock.owner == current_thread:
            lock.entry_count += 1
            return True
    elif not lock.pending_requests:
        # not locked, not requested - go!
        lock.owner = current_thread
        lock.entry_count = 1
        return True
    # need to get the real lock
    return _acquire_lock(lock, current_thread, blocking)


###########################################################################
## public C-API

cdef create_fastrlock():
    """
    Public C level entry function for creating a FastRlock instance.
    """
    return FastRLock.__new__(FastRLock)


cdef bint lock_fastrlock(rlock, long current_thread, bint blocking) except -1:
    """
    Public C level entry function for locking a FastRlock instance.

    The 'current_thread' argument is deprecated and ignored.  Pass -1 for backwards compatibility.
    """
    # Note: 'current_thread' used to be set to -1 or the current thread ID, but -1 is signed while "pythread_t" isn't.
    return _lock_rlock(&(<FastRLock?>rlock)._real_lock, pythread.PyThread_get_thread_ident(), blocking)


cdef int unlock_fastrlock(rlock) except -1:
    """
    Public C level entry function for unlocking a FastRlock instance.
    """
    _unlock_lock(&(<FastRLock?>rlock)._real_lock)
    return 0




from cpython cimport pythread

from fastrlock import LockNotAcquired

cdef extern from *:
    # Compatibility definitions for Python
    """
    #if PY_VERSION_HEX >= 0x030700a2
    typedef unsigned long pythread_t;
    #else
    typedef long pythread_t;
    #endif
    """

    # Just let Cython understand that pythread_t is
    # a long type, but be aware that it is actually
    # signed for versions of Python prior to 3.7.0a2 and
    # unsigned for later versions
    ctypedef unsigned long pythread_t


cdef struct _LockStatus:
    pythread.PyThread_type_lock lock
    pythread_t owner               # thread ID of the current lock owner
    unsigned int entry_count       # number of (re-)entries of the owner
    unsigned int pending_requests  # number of pending requests for real lock
    bint is_locked                 # whether the real lock is acquired


cdef bint _acquire_lock(_LockStatus *lock, long current_thread,
                        bint blocking) nogil except -1:
    # Note that this function *must* hold the GIL when being called.
    # We just use 'nogil' in the signature to make sure that no Python
    # code execution slips in that might free the GIL

    wait = pythread.WAIT_LOCK if blocking else pythread.NOWAIT_LOCK
    if not lock.is_locked and not lock.pending_requests:
        # someone owns it but didn't acquire the real lock - do that
        # now and tell the owner to release it when done
        if pythread.PyThread_acquire_lock(lock.lock, pythread.NOWAIT_LOCK):
            lock.is_locked = True
    #assert lock._is_locked

    lock.pending_requests += 1
    # wait for the lock owning thread to release it
    with nogil:
        while True:
            locked = pythread.PyThread_acquire_lock(lock.lock, wait)
            if locked:
                break
            if wait == pythread.NOWAIT_LOCK:
                lock.pending_requests -= 1
                return False
    lock.pending_requests -= 1
    #assert not lock.is_locked
    #assert lock.reentry_count == 0
    #assert locked
    lock.is_locked = True
    lock.owner = current_thread
    lock.entry_count = 1
    return True


cdef inline void _unlock_lock(_LockStatus *lock) nogil noexcept:
    # Note that this function *must* hold the GIL when being called.
    # We just use 'nogil' in the signature to make sure that no Python
    # code execution slips in that might free the GIL

    #assert lock.entry_count > 0
    lock.entry_count -= 1
    if lock.entry_count == 0:
        if lock.is_locked:
            pythread.PyThread_release_lock(lock.lock)
            lock.is_locked = False
